# normal data
# augmented sampling scheme

# assumes existence of:
# y: a vector of observations
# locns: a 2-column matrix of locations of observations with same number of
#    rows as length of y
# locns.test: a 2-column matrix of locations of prediction locations with
#    same number of rows as length of y

useJointEta2Sampling <- FALSE
# flag for whether to jointly sample noise variance('eta2') and augmented data

library(spectralGP)

nBurnin <- 1000 
nIts <- 10000 

k <- 128
# size of grid in each dimension; k / 2 + 1 grid cells are actually used
#   for the domain of interest
half <- k / 2 + 1

set.seed(0)

mirror <- function(mat){
# reflects value across midpoint or midlines of domain for use in initializing
#   process outside domain of interest
  if(d == 2){
    mat[1:half, (half + 1):k] <- mat[1:half, (half - 1):2]
    mat[(half + 1):k, ] <- mat[(half - 1):2, ]
  } else{
    mat[(half + 1):k] <- mat[(half - 1):2]
  }
  return(mat)
}

########################################################################################
# for real data that aren't already in (0, 1)^2, need to do mapping to unit domain here;
# otherwise keep next lines commented out
########################################################################################
if(FALSE){
  locns.orig <- locns
  locns <- xy2unit(locns, locations.scale = rbind(locns, locns.test))
  locns.test.orig <- locns.test
  locns.test <- xy2unit(locns.test, locations.scale = rbind(locns, locns.test))
}

########################################################################
# set up initial values and data structures
########################################################################

# set up mapping of locations to grid
fun <- gp(c(k, k))
map.train <- c(new.mapping(fun, locns))
n <- length(locns[ ,1])
map.test <- c(new.mapping(fun, locns.test))

n.gridded <- rep(0, k * k)
tmp <- table(map.train)
n.gridded[as.numeric(names(tmp))] <- tmp  
n.filled <- n.gridded[n.gridded > 0]
nUnfilled <- sum(n.gridded == 0)
nFilled <- sum(n.gridded > 0)

# set up mapping from grid to observation locations
indic <- (1:(k * k)) * (n.gridded > 0)
indic <- indic[indic > 0]
invMap.train <- matrix(0, nr = n, nc = nFilled)
for(i in 1:nFilled){
  invMap.train[map.train == indic[i], i] <- 1
}
invMap.train <- t(invMap.train)

# determine which grid cells have data
filledCells <- (1:(k * k)) * (n.gridded > 0)
filledCells <- filledCells[filledCells > 0]
unfilledCells <- (1:(k * k)) * (n.gridded == 0)
unfilledCells <- unfilledCells[unfilledCells > 0]
y.filled <- invMap.train %*% y / n.filled
y.gridded <- rep(0, k * k)
y.gridded[filledCells] <- y.filled

# use gam() to get initial fit
dat <- data.frame(y = y, x1 = locns[ ,1], x2 = locns[ ,2])
mod <- gam(y~s(x1, x2, k = min(n - 5, 500)), dat = dat)
fullGrid <- expand.gpgrid(fun)

# calculate initial values at grid cells
pred.griddedVec <- predict(mod, newdata = list(x1 = fullGrid[ , 1], x2 = fullGrid[ ,2])) + 
  rnorm(half * half, 0, sqrt(mod$sig2))
pred.griddedMat <- matrix(0, nr = k, nc = k)
pred.griddedMat[1:(half), 1:(half)] <- matrix(pred.griddedVec, half, half)
pred.griddedVec <- c(pred.griddedMat)
pred.griddedVec[filledCells] <- y.filled
pred.griddedMat <- matrix(pred.griddedVec, k, k)
pred.griddedVec <- c(mirror(pred.griddedMat))

#####################################
### set up parameters and processes
#####################################

param <- function(initial = 0, transformFxn = exp, name = NULL, priorParameters = NULL, 
  priorRange <- c(-Inf, Inf)){

  # function to initialize parameter objects
  object  <-  list()
  object$name <- name

  object$value <- transformFxn(initial)
  object$baseValue <- initial
  object$p <- length(initial)
  object$transformFxn <- transformFxn

  object$priorParameters <- priorParameters
  object$priorRange <- priorRange

  class(object) <- 'param'
  return(object)
}

identity <- function(x){x}

decide <- function(post.logdiff){
  # Metropolis-Hastings acceptance decision
  n <- length(post.logdiff)
  rdev <- runif(n)
  accepted <- rep(TRUE, n)
  accepted[exp(post.logdiff) <= rdev] <- FALSE
  return(accepted)
}

# noise variance
if(!useJointEta2Sampling){
  eta2 <- param(mod$sig2, identity, name = 'eta2', priorParameters = c(.1, .1), 
    priorRange = c(0.0001,100))
} else{
# if jointly sampling eta2 and the augmented observations, I suggest
#   using a log normal prior
  eta2 <- param(log(mod$sig2), exp, name = 'eta2', priorParameters = c(0,9), 
    priorRange = c(-10,10))
}

# process standard deviation
gamma <- param(log(sd(y)), exp, name = 'gamma', priorParameters = c(0,9), 
  priorRange = c(-10,10))

# process range
rho <- param(log(0.05), exp, name = 'rho', priorParameters = c(0,10), 
  priorRange = c(log(.001),log(5)))

# process mean
mu <- param(mean(y), identity, name = 'mu', priorParameters = c(0,10), 
  priorRange = c(-Inf,Inf))

# process smoothness
nu <- param(4, identity, name = 'nu', priorParameters = c(0.5, 15), 
  priorRange = c(0.5,15))

# initialize spatial process
fun <- gp(c(k, k), matern.specdens, c(rho$value, nu$value), 1, const.fixed = TRUE)
fun$logdensity <- logdensity(fun)
fun$name <- 'fun'

# initialize augmented data vector
yFull <- param(pred.griddedVec, identity, name = 'yFull')

#####################################
### prior and likelihood functions
#####################################

# this is used only if doing joint sampling of eta2 and augmented data
logLik <- function(){
  # loglikelihood of observed data
  return(sum(dnorm(y, mu$value + gamma$value * fun$process[map.train], 
                   sqrt(eta2$value), log = TRUE)))
}

logLikFull <- function(){
  # loglikelihood of augmented data
  return(sum(dnorm(yFull$value, mu$value + gamma$value * fun$process, sqrt(eta2$value), log = TRUE)))
}

logLikFullFilled <- function(){
  # loglikelihood at grid cells with observed data
  # (based on averaging observations in each grid cell)
  return(sum(dnorm(yFull$value[filledCells], mu$value + gamma$value * 
                   fun$process[filledCells], sqrt(eta2$value), log = TRUE)))
}

yFull.Gibbs <- function(){
  # Gibbs sample for augmented data
  newSample <- rnorm(k * k, mu$value + gamma$value * fun$process, sqrt(eta2$value))
  newSample[filledCells] <- yFull$value[filledCells]
  return(c((matrix(newSample, nr = k, nc = k))))
}

mu.Gibbs <- function(){
  v <- 1 / ((k * k / eta2$value + 1 / (mu$priorParameters[2]^2)))
  rnorm(1, v * (sum((yFull$value - gamma$value * fun$process)) / eta2$value + 
             mu$priorParameters[1] / (mu$priorParameters[2]^2)), sqrt(v))
}

gammaPrior <- function(){
  if(max(gamma$baseValue) > gamma$priorRange[2] | min(gamma$baseValue) <
     gamma$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(gamma$baseValue, gamma$priorParameters[1], 
                 gamma$priorParameters[2], log = TRUE))
  }
}

eta2.Gibbs <- function(){
  return(1 / rgamma(1, k * k / 2 + eta2$priorParameters[1], rate = eta2$priorParameters[2] + 
                  0.5 * sum((yFull$value - mu$value - gamma$value * fun$process)^2)))
}

if(!useJointEta2Sampling){
  eta2Prior <- function(){
    if(max(eta2$baseValue) > eta2$priorRange[2] | min(eta2$baseValue) < 
       eta2$priorRange[1]){
      return(-Inf)
    } else{
      return(-(eta2$priorParameters[1] + 1) * log(eta2$baseValue) - 
             eta2$priorParameters[2] / eta2$baseValue)
    }
  }
} else{
# if using joint sampling of eta2 and augmented observations, 
#  makes use of the following prior function for eta2 rather than the code above
  eta2Prior <- function(){
    if(max(eta2$baseValue) > eta2$priorRange[2] | min(eta2$baseValue) < 
       eta2$priorRange[1]){
      return(-Inf)
    } else{
      return(dnorm(eta2$baseValue, eta2$priorParameters[1], 
                   eta2$priorParameters[2], log = TRUE))
    }
  }
}

muPrior <- function(){
  if(max(mu$baseValue) > mu$priorRange[2] | min(mu$baseValue) < mu$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(mu$baseValue, mu$priorParameters[1], mu$priorParameters[2], log = TRUE))
  }
}


rhoPrior <- function(){
  if(max(rho$baseValue) > rho$priorRange[2] | min(rho$baseValue) < rho$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(rho$baseValue, rho$priorParameters[1], rho$priorParameters[2], log = TRUE))
  }
}


funPrior <- function(){
  logdensity(fun)
}

# proposal standard deviations for Metropolis sampling
rho$proposalSD <- 0.03
gamma$proposalSD <- 0.03
if(useJointEta2Sampling){
  eta2$proposalSD <- 0.03 
}

totaln <- nBurnin + nIts
cat('Starting iterations: ', date(), '\n')

# initialize coefficients and augmented data
Gibbs.sample.coeff(fun, yFull$value, eta2$value, mu$value, gamma$value)
yFull$value <- yFull$baseValue <- yFull.Gibbs()

# sampling iterations
for(it in 1:totaln){

  # sample coefficients
  Gibbs.sample.coeff(fun, yFull$value, eta2$value, mu$value, gamma$value)

  # sample augmented data
  yFull$value <- yFull$baseValue <- yFull.Gibbs()

  # sample noise variance either using Gibbs sample or with
  #  joint sample with augmented data
  if(!useJointEta2Sampling){
    eta2$value <- eta2.Gibbs()
  } else{
    eta2Save <- eta2
    lpValue <- eta2Prior() + logLikFullFilled()
    eta2$baseValue <- rnorm(1, eta2$baseValue, eta2$proposalSD) 
    eta2$value <- eta2$transformFxn(eta2$baseValue)
    lpProp <- eta2Prior() + logLikFullFilled()
    accepted <- decide(lpProp - lpValue)
    if(accepted){
      yFull$value <- yFull$baseValue <- yFull.Gibbs()
      # sample augmented data if accept joint proposal
    } else{
      eta2 <- eta2Save
    }
  }

  # sample process standard deviation
  gammaSave <- gamma
  lpValue <- gammaPrior() + logLikFull()
  gamma$baseValue <- rnorm(1, gamma$baseValue, gamma$proposalSD)
  gamma$value <- gamma$transformFxn(gamma$baseValue)
  lpProp <- gammaPrior() + logLikFull()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    gamma <- gammaSave
  }
  
  # sample process range
  rhoSave <- rho
  lpValue <- rhoPrior() + funPrior()
  rho$baseValue <- rnorm(1, rho$baseValue, rho$proposalSD)
  rho$value <- rho$transformFxn(rho$baseValue)
  funSave <- copy(fun)
  change.param(fun, c(rho$value, nu$value))
  lpProp <- rhoPrior() + funPrior()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    copy(funSave, fun)
    rho <- rhoSave
  }

  # sample process mean
  mu$value <- mu.Gibbs()

  # calculate process values at training, test, and all grid locations
  f.train <- mu$value + gamma$value * predict(fun, mapping = map.train)
  f.test <- mu$value + gamma$value * predict(fun, mapping = map.test)
  f.grid <- mu$value + gamma$value * predict(fun)

}
