# exponential family data
# sampling scheme follows Wikle (2002)

# assumes existence of:
# y: a vector of observations
# locns: a 2-column matrix of locations of observations with same
#   number of rows as length of y
# locns.test: a 2-column matrix of locations of prediction locations
#   with same number of rows as length of y
# m (for binomial), a vector of #trials

library(spectralGP)

if(family == 'poisson'){
  linkFxn <- log
  invLinkFxn <- exp
}
if(family == 'binomial'){
  linkFxn <- f.logit
  invLinkFxn <- f.invlogit
}

nBurnin <- 1000 
nIts <- 10000 

k <- 128
# size of grid in each dimension; k / 2+1 grid cells are
#  actually used for the domain of interest
half <- k / 2+1

eta2.init <- 0.1^2
tau2.init <- 0.1^2

set.seed(0)

mirror <- function(mat){
# reflects value across midpoint or midlines of domain for use in
#   initializing process outside domain of interest
  if(d == 2){
    mat[1:half, (half+1):k] <- mat[1:half, (half - 1):2]
    mat[(half+1):k, ] <- mat[(half - 1):2, ]
  } else{
    mat[(half+1):k] <- mat[(half - 1):2]
  }
  return(mat)
}

########################################################################################
# for real data that aren't already in (0,1)^2, need to do mapping to unit domain here;
# otherwise keep next lines commented out
########################################################################################
if(FALSE){
  locns.orig <- locns
  locns <- xy2unit(locns, locations.scale = rbind(locns, locns.test))
  locns.test.orig <- locns.test
  locns.test <- xy2unit(locns.test, locations.scale = rbind(locns, locns.test))
}

# set up mapping of locations to grid
fun <- gp(c(k, k))
map.train <- c(new.mapping(fun, locns))
n <- length(locns[ , 1])
map.test <- c(new.mapping(fun, locns.test))

n.gridded <- rep(0, k * k)
tmp <- table(map.train)
n.gridded[as.numeric(names(tmp))] <- tmp  # same as KtK
n.filled <- n.gridded[n.gridded > 0]
nUnfilled <- sum(n.gridded == 0)
nFilled <- sum(n.gridded > 0)

# set up mapping from grid to observation locations
indic <- (1:(k * k)) * (n.gridded > 0)
indic <- indic[indic > 0]
invMap.train <- matrix(0, nr = n, nc = nFilled)
for(i in 1:nFilled){
  invMap.train[map.train == indic[i], i] <- 1
}
invMap.train <- t(invMap.train)

# determine which grid cells have data
filledCells <- (1:(k * k)) * (n.gridded > 0)
filledCells <- filledCells[filledCells > 0]
unfilledCells <- (1:(k * k)) * (n.gridded == 0)
unfilledCells <- unfilledCells[unfilledCells > 0]
if(family == 'poisson'){
  y.filled <- invMap.train %*% y / n.filled
}
if(family == 'binomial'){
  y.filled <- (invMap.train %*% y) / (invMap.train %*% m)
}  
y.gridded <- rep(0, k*k)
y.gridded[filledCells] <- y.filled

# use gam() to get initial fit
dat <- data.frame(y = y, x1 = locns[ , 1], x2 = locns[ ,2])
mod <- gam(y~s(x1, x2, k = min(n - 5, 500)), dat = dat, family = family)
fullGrid <- expand.gpgrid(fun)

# calculate initial values at grid cells
pred.griddedVec <- predict(mod, newdata = list(x1 = fullGrid[ , 1], x2 = fullGrid[ , 2]),
                           type = 'link') + rnorm(half * half, 0, sqrt(eta2.init))
pred.griddedMat <- matrix(0, nr = k, nc = k)
pred.griddedMat[1:(half), 1:(half)] <- matrix(pred.griddedVec, half, half)
pred.griddedVec <- c(mirror(pred.griddedMat))

#####################################
### set up parameters and processes
#####################################

param <- function(initial = 0, transformFxn = exp, name = NULL,
                  priorParameters = NULL, priorRange = c(-Inf, Inf)){
  # function to initialize parameter objects
  object  <-  list()
  object$name <- name

  object$value <- transformFxn(initial)
  object$baseValue <- initial
  object$p <- length(initial)
  object$transformFxn <- transformFxn

  object$priorParameters <- priorParameters
  object$priorRange <- priorRange

  class(object) <- 'param'
  return(object)
}

identity <- function(x){x}

decide <- function(post.logdiff){
  # Metropolis-Hastings acceptance decision
  n <- length(post.logdiff)
  rdev <- runif(n)
  accepted <- rep(TRUE, n)
  accepted[exp(post.logdiff) <= rdev] <- FALSE
  return(accepted)
}

# variance components
eta2 <- param(eta2.init, identity, name = 'eta2', priorParameters = c(.5, .5),
              priorRange = c(0.0001, 100))
sig2z <- param(sig2z.init, identity, name = 'sig2z', priorParameters = c(1, .1),
               priorRange = c(0.0001, 100))

# process standard deviation
gamma <- param(log(sd(y)), exp, name = 'gamma', priorParameters = c(0, 9),
               priorRange = c(-10, 10))

# process range
rho <- param(log(0.05), exp, name = 'rho', priorParameters = c(0, 10),
             priorRange = c(log(.001), log(5)))

# process mean
mu <- param(linkFxn(mean(y)), identity, name = 'mu', priorParameters = c(0, 10),
            priorRange = c(-Inf, Inf))

# process smoothness
nu <- param(4, identity, name = 'nu', priorParameters = c(0.5, 15),
            priorRange = c(0.5, 15))

# initialize spatial process
fun <- gp(c(k, k), matern.specdens, c(rho$value, nu$value), 1, const.fixed = TRUE)
fun$logdensity <- logdensity(fun)
fun$name <- 'fun'

# 'z' process
z <- param((pred.griddedVec - mu$value) / gamma$value, identity, name = 'z',
           priorParameters = c(0, 100), priorRange = c(-50, 50))

# 'lambda' process; one lambda value for each observation
lambda <- param(rnorm(n, mu$value + gamma$value * z$value[map.train], sqrt(eta2$value)),
                identity, name = 'lambda', priorParameters = c(0, 100),
                priorRange = c(-50, 50))  

# initialize process
Gibbs.sample.coeff(fun, z$value, sig2z$value, 0, 1)

#####################################
### prior and likelihood functions
#####################################

logLik <- function(){
  if(family == 'poisson'){
    return(sum(dpois(y, invLinkFxn(lambda$value), log = TRUE)))
  }
  if(family == 'binomial'){
    return(sum(dbinom(y, m, invLinkFxn(lambda$value), log = TRUE)))
  }
  stop('Family (Poisson or binomial) not specified')
}

# loglikelihood for each observation
logLikVec <- function(){
  if(family == 'poisson'){
    return(dpois(y, invLinkFxn(lambda$value), log = TRUE))
  }
  if(family == 'binomial'){
    return(dbinom(y, m, invLinkFxn(lambda$value), log = TRUE))
  }
  stop('Family (Poisson or binomial) not specified')
}  

lambdaPrior <- function(){
  priorValue <- dnorm(lambda$value, mu$value + gamma$value * z$value[map.train],
                      sqrt(eta2$value), log = TRUE)
  priorValue[lambda$value < lambda$priorRange[1] |
             lambda$value > lambda$priorRange[2]] <- -Inf
  return(sum(priorValue))
}

# lambda prior for each lambda value
lambdaPriorVec <- function(){
  priorValue <- dnorm(lambda$value, mu$value + gamma$value * z$value[map.train],
                      sqrt(eta2$value), log = TRUE)
  priorValue[lambda$value < lambda$priorRange[1] |
             lambda$value > lambda$priorRange[2]] <- -Inf
  return(priorValue)  
}

z.Gibbs <- function(){
  zVar <- rep(sig2z$value, k * k)
  zVar[filledCells] <- 1 / (1 / sig2z$value + gamma$value * gamma$value *
                            n.filled / eta2$value)
  zMean <- fun$process / sig2z$value
  zMean[filledCells] <- zMean[filledCells] + gamma$value * invMap.train %*%
    (lambda$value - mu$value) / eta2$value
  zMean <- zVar * zMean
  return(rnorm(k * k, zMean, sqrt(zVar)))
}

###########################################
# not used but included for completeness
zPrior <- function(){
  priorValue <- dnorm(fun$process, sqrt(sig2z$value), log = TRUE)
  priorValue[z$value < z$priorRange[1] | z$value > z$priorRange[2]] <- -Inf
  return(sum(priorValue))
}
zPriorFilled <- function(){
  priorValue <- dnorm(z$value[filledCells], fun$process[filledCells],
                      sqrt(sig2z$value), log = TRUE)
  priorValue[z$value[filledCells] < z$priorRange[1] | z$value[filledCells] >
             z$priorRange[2]] <- -Inf
  return(sum(priorValue))
}
##############################################

mu.Gibbs <- function(){
  v <- 1 / ((n / eta2$value + 1 / (mu$priorParameters[2]^2)))
  rnorm(1, v * (sum((lambda$value - gamma$value * z$value[map.train])) / eta2$value
                + mu$priorParameters[1] / (mu$priorParameters[2]^2)), sqrt(v))
}

gammaPrior <- function(){
  if(max(gamma$baseValue) > gamma$priorRange[2] | min(gamma$baseValue) <
     gamma$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(gamma$baseValue, gamma$priorParameters[1],
                 gamma$priorParameters[2], log = TRUE))
  }
}

muPrior <- function(){
  if(max(mu$baseValue) > mu$priorRange[2] | min(mu$baseValue) < mu$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(mu$baseValue, mu$priorParameters[1],
                 mu$priorParameters[2], log = TRUE))
  }
}


rhoPrior <- function(){
  if(max(rho$baseValue) > rho$priorRange[2] | min(rho$baseValue) < rho$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(rho$baseValue, rho$priorParameters[1],
                 rho$priorParameters[2], log = TRUE))
  }
}

eta2.Gibbs <- function(){
  return(1 / rgamma(1, n / 2 + eta2$priorParameters[1],
                    rate = eta2$priorParameters[2] + 0.5 *
                    sum((lambda$value - mu$value - gamma$value *
                         z$value[map.train])^2)))
}

eta2Prior <- function(){
  if(max(eta2$baseValue) > eta2$priorRange[2] | min(eta2$baseValue) <
     eta2$priorRange[1]){
    return(-Inf)
  } else{
    return(-(eta2$priorParameters[1] + 1) * log(eta2$baseValue) -
           eta2$priorParameters[2] / eta2$baseValue)
  }
}
  
sig2zPrior <- function(){
  if(max(sig2z$baseValue) > sig2z$priorRange[2] | min(sig2z$baseValue) <
     sig2z$priorRange[1]){
    return(-Inf)
  } else{
    return(-(sig2z$priorParameters[1] + 1) * log(sig2z$baseValue) -
           sig2z$priorParameters[2] / sig2z$baseValue)
  }
}


sig2z.Gibbs <- function(){
  return(1 / rgamma(1, (k * k) / 2 + sig2z$priorParameters[1],
                    rate = sig2z$priorParameters[2] + 
                  0.5 * sum((z$value - fun$process)^2)))
}

funPrior <- function(){
  logdensity(fun)
}

# proposal standard deviations for Metropolis sampling
lambda$proposalSD <- 0.03
gamma$proposalSD <- 0.03
rho$proposalSD <- 0.03



totaln <- nBurnin + nIts
cat('Starting iterations: ', date(), '\n')
for(it in 1:totaln){

  # sample coefficients
  Gibbs.sample.coeff(fun, z$value, sig2z$value, 0, 1)

  # sample 'z' process
  z$value <- z.Gibbs()

  # sample 'lambda' process with individual acceptance for each element
  lambdaSave <- lambda
  lpValue <- lambdaPriorVec() + logLikVec()
  lambda$value <- rnorm(n, lambda$value, lambda$proposalSD)
  lpProp <- lambdaPriorVec() + logLikVec()
  accepted <- decide(lpProp - lpValue)
  accepted[lambda$value < lambda$priorRange[1] | lambda$value >
           lambda$priorRange[2]] <- FALSE
  lambda$value[!accepted] <- lambdaSave$value[!accepted]

  # sample z variance component
  sig2z$value <- sig2z.Gibbs()

  # sample process standard deviation
  gammaSave <- gamma
  lpValue <- gammaPrior() + lambdaPrior()
  gamma$baseValue <- rnorm(1, gamma$baseValue, gamma$proposalSD)
  gamma$value <- gamma$transformFxn(gamma$baseValue)
  lpProp <- gammaPrior() + lambdaPrior()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    gamma <- gammaSave
  }

  # sample process range
  rhoSave <- rho
  lpValue <- rhoPrior() + funPrior()
  rho$baseValue <- rnorm(1, rho$baseValue, rho$proposalSD)
  rho$value <- rho$transformFxn(rho$baseValue)
  funSave <- copy(fun)
  change.param(fun, c(rho$value, nu$value))
  lpProp <- rhoPrior() + funPrior()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    copy(funSave, fun)
    rho <- rhoSave
  }

  # sample process mean
  mu$value <- mu.Gibbs()

  # sample lambda variance component
  eta2$value <- eta2.Gibbs()

  # calculate process values at training,  test,  and all grid locations
  # note that this uses smoothest version of process based on GP
  f.train <- mu$value + gamma$value * predict(fun, mapping  =  map.train) 
  f.test <- mu$value + gamma$value * predict(fun, mapping = map.test)
  f.grid <- mu$value + gamma$value * predict(fun)

}


