# exponential family data
# code for blocked sampling
# joint sampling of rho (process range) & sigma (process standard deviation) with
#    process coefficients
# (uses representation with sigma rather than gamma)
# within the joint sample, the process coefficients are proposed deterministically

# assumes existence of:
# y: a vector of observations
# locns: a 2-column matrix of locations of observations with same number of rows as length of y
# locns.test: a 2-column matrix of locations of prediction locations with same number of
#   rows as length of y
# m (for binomial), a vector of #trials


library(spectralGP)

if(family == 'poisson'){
  linkFxn <- log
  invLinkFxn <- exp
}
if(family == 'binomial'){
  linkFxn <- f.logit
  invLinkFxn <- f.invlogit
}

nBurnin <- 1000 
nIts <- 10000 

k <- 128
# size of grid in each dimension; k/2+1 grid cells are actually used for the domain of interest
half <- k / 2 + 1

addedInitNoise <- 0.3^2  # noise variance used solely to initialize coefficients via a Gibbs sample

set.seed(0)

mirror <- function(mat){
# reflects value across midpoint or midlines of domain for use in initializing process
  #  outside domain of interest
  if(d == 2){
    mat[1:half, (half + 1):k] <- mat[1:half, (half - 1):2]
    mat[(half + 1):k, ] <- mat[(half - 1):2, ]
  } else{
    mat[(half + 1):k] <- mat[(half - 1):2]
  }
  return(mat)
}

########################################################################################
# for real data that aren't already in (0,1)^2, need to do mapping to unit domain here;
# otherwise keep next lines commented out
########################################################################################
if(FALSE){
  locns.orig <- locns
  locns <- xy2unit(locns, locations.scale = rbind(locns, locns.test))
  locns.test.orig <- locns.test
  locns.test <- xy2unit(locns.test, locations.scale = rbind(locns, locns.test))
}

# set up mapping of locations to grid
fun <- gp(c(k, k))
map.train <- c(new.mapping(fun, locns))
n <- length(locns[ ,1])
map.test <- c(new.mapping(fun, locns.test))

n.gridded <- rep(0, k * k)
tmp <- table(map.train)
n.gridded[as.numeric(names(tmp))] <- tmp  # same as KtK
n.filled <- n.gridded[n.gridded > 0]
nUnfilled <- sum(n.gridded == 0)
nFilled <- sum(n.gridded > 0)

# set up mapping from grid to observation locations
indic <- (1:(k * k)) * (n.gridded > 0)
indic <- indic[indic > 0]
invMap.train <- matrix(0, nr = n, nc = nFilled)
for(i in 1:nFilled){
  invMap.train[map.train == indic[i], i] <- 1
}
invMap.train <- t(invMap.train)

filledCells <- (1:(k * k)) * (n.gridded > 0)
filledCells <- filledCells[filledCells > 0]
unfilledCells <- (1:(k * k)) * (n.gridded == 0)
unfilledCells <- unfilledCells[unfilledCells > 0]
if(family == 'poisson'){
  y.filled <- invMap.train %*% y / n.filled
}
if(family == 'binomial'){
  y.filled <- (invMap.train %*% y) / (invMap.train %*% m)
}  
y.gridded <- rep(0, k*k)
y.gridded[filledCells] <- y.filled

# use gam() to get initial fit
dat <- data.frame(y = y, x1 = locns[ , 1], x2 = locns[ ,2])
mod <- gam(y~s(x1, x2, k = min(n - 5, 500)), dat = dat, family = family)
fullGrid <- expand.gpgrid(fun)

# calculate initial values at grid cells
pred.griddedVec <- predict(mod, newdata = list(x1 = fullGrid[ , 1], x2 = fullGrid[ ,2]),
                           type = 'link') + rnorm(half * half, 0, sqrt(addedInitNoise))
pred.griddedMat <- matrix(0, nr = k, nc = k)
pred.griddedMat[1:(half), 1:(half)] <- matrix(pred.griddedVec, half, half)
pred.griddedVec <- c(mirror(pred.griddedMat))


#####################################
### set up parameters and processes
#####################################

param <- function(initial = 0, transformFxn = exp, name = NULL, priorParameters = NULL,
                  priorRange = c(-Inf, Inf)){
  # function to initialize parameter objects
  object <- list()
  object$name <- name

  object$value <- transformFxn(initial)
  object$baseValue <- initial
  object$p <- length(initial)
  object$transformFxn <- transformFxn

  object$priorParameters <- priorParameters
  object$priorRange <- priorRange

  class(object) <- 'param'
  return(object)
}

identity <- function(x){x}

decide <- function(post.logdiff){
  # Metropolis-Hastings acceptance decision
  n <- length(post.logdiff)
  rdev <- runif(n)
  accepted <- rep(TRUE, n)
  accepted[exp(post.logdiff) <= rdev] <- FALSE
  return(accepted)
}

# process standard deviation
sigma <- param(log(sd(y)), exp, name = 'sigma', priorParameters = c(0, 9), priorRange = c(-10, 10))

# process range
rho <- param(log(0.05), exp, name = 'rho', priorParameters = c(0, 10), priorRange =
             c(log(.01), log(5)))

# process mean
mu <- param(linkFxn(mean(y)), identity, name = 'mu', priorParameters = c(0, 10),
            priorRange = c(-Inf, Inf))

# process smoothness
nu <- param(4, identity, name = 'nu', priorParameters = c(0.5, 15), priorRange = c(0.5, 15))

# initialize spatial process
fun <- gp(c(k, k), matern.specdens, c(rho$value, nu$value), sigma$value^2, const.fixed = TRUE)
fun$logdensity <- logdensity(fun)
fun$name <- 'fun'
add.blocks(fun)
fun$proposalSD <- c(.15, .2, .2, .15, .05, .04, .02, .01, .005, .0025)
fun$proposalSD <- fun$proposalSD[1:fun$num.blocks]

Gibbs.sample.coeff(fun, pred.griddedVec, addedInitNoise, mu$value, 1)


#####################################
### prior and likelihood functions
#####################################

logLik <- function(){
  if(family == 'poisson'){
    return(sum(dpois(y, invLinkFxn(mu$value + predict(fun, mapping = map.train)), log = TRUE)))
  }
  if(family == 'binomial'){
    return(sum(dbinom(y, m, invLinkFxn(mu$value + predict(fun, mapping = map.train)),
                      log = TRUE)))
  }
}

muPrior <- function(){
  if(max(mu$baseValue) > mu$priorRange[2] | min(mu$baseValue) < mu$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(mu$baseValue, mu$priorParameters[1], mu$priorParameters[2], log = TRUE))
  }
}


rhoPrior <- function(){
  if(max(rho$baseValue) > rho$priorRange[2] | min(rho$baseValue) < rho$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(rho$baseValue, rho$priorParameters[1], rho$priorParameters[2], log = TRUE))
  }
}

sigmaPrior <- function(){
  if(max(sigma$baseValue) > sigma$priorRange[2] | min(sigma$baseValue) < sigma$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(sigma$baseValue, sigma$priorParameters[1], sigma$priorParameters[2], log = TRUE))
  }
}

funPrior <- function(){
  logdensity(fun)
}


# proposal standard deviations for Metropolis sampling
sigma$proposalSD <- 0.03
rho$proposalSD <- 0.03
mu$proposalSD <- 0.03

totaln <- nBurnin + nIts
cat('Starting iterations: ',date(),'\n')

for(it in 1:totaln){
  
  # sample coefficients in blocks
  lpValue <- logLik() + funPrior()
  for(block in 1:fun$num.blocks){  # propose each block of coefficients
    if(fun$proposalSD[block]){
      funSave <- copy(fun)
      propose.coeff(fun, block, fun$proposalSD[block])
      lpProp <- logLik() + funPrior()
      accepted <- decide(lpProp - lpValue)
      if(accepted){
        lpValue <- lpProp
      } else{
	copy(funSave, fun)
      }
    }
  }

  # sample process standard deviation jointly with coefficients using deterministic
  #   move of coefficients
  funSave <- copy(fun)
  sigmaSave <- sigma
  lpValue <- sigmaPrior() + logLik()
  sigma$baseValue <- rnorm(1, sigma$baseValue, sigma$proposalSD)
  sigma$value <- sigma$transformFxn(sigma$baseValue)
  change.param(fun, new.variance.param = sigma$value^2)
  fun$coeff <- fun$coeff * sqrt(fun$variances / funSave$variances)
        # deterministic shift of coefficients
  updateprocess(fun)
  lpProp <- sigmaPrior() + logLik()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    copy(funSave, fun)
    sigma <- sigmaSave
  }

  # sample process range jointly with coefficients using deterministic move of coefficients
  rhoSave <- rho
  lpValue <- rhoPrior() + logLik()
  rho$baseValue <- rnorm(1, rho$baseValue, rho$proposalSD)
  rho$value <- rho$transformFxn(rho$baseValue)
  funSave <- copy(fun)
  change.param(fun, c(rho$value, nu$value))
  fun$coeff <- fun$coeff * sqrt(fun$variances / funSave$variances)
        # deterministic shift of coefficients
  updateprocess(fun)
  lpProp <- rhoPrior() + logLik()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    copy(funSave, fun)
    rho <- rhoSave
  }

  # sample process mean
  muSave <- mu
  lpValue <- muPrior() + logLik()
  mu$baseValue <- rnorm(1, mu$baseValue, mu$proposalSD)
  mu$value <- mu$transformFxn(mu$baseValue)
  lpProp <- muPrior() + logLik()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    mu <- muSave
  }

  # calculate process values at training,  test,  and all grid locations
  f.train <- mu$value + predict(fun, mapping = map.train)
  f.test <- mu$value + predict(fun, mapping = map.test)
  f.grid <- mu$value + predict(fun)
}  


