# exponential family data
# modified Wikle sampling scheme
# joint sampling of psi1 and psi2 (transformations of rho & sigma) with process
#   coefficients and lambda process values
# within the joint sample, the process coefficients and lambda values are proposed deterministically

# assumes existence of:
# y: a vector of observations
# locns: a 2-column matrix of locations of observations with same number of rows as length of y
# locns.test: a 2-column matrix of locations of prediction locations with same number of rows
#    as length of y
# m (for binomial), a vector of #trials

useJointEta2Sampling <- FALSE
# flag for whether to jointly sample lambda variance ('eta2') and lambda process values 


library(spectralGP)

if(family == 'poisson'){
  linkFxn <- log
  invLinkFxn <- exp
}
if(family == 'binomial'){
  linkFxn <- f.logit
  invLinkFxn <- f.invlogit
}

nBurnin <- 1000 
nIts <- 10000 

k <- 128  # size of grid in each dimension; k/2+1 grid cells are actually used for the domain of interest
half <- k / 2 + 1

eta2.init <- 0.1^2

set.seed(0)

mirror <- function(mat){
# reflects value across midpoint or midlines of domain for use in initializing process outside
  #  domain of interest
  if(d == 2){
    mat[1:half, (half + 1):k] <- mat[1:half, (half - 1):2]
    mat[(half + 1):k, ] <- mat[(half - 1):2, ]
  } else{
    mat[(half + 1):k] <- mat[(half - 1):2]
  }
  return(mat)
}

########################################################################################
# for real data that aren't already in (0,1)^2, need to do mapping to unit domain here;
# otherwise keep next lines commented out
########################################################################################
if(FALSE){
  locns.orig <- locns
  locns <- xy2unit(locns, locations.scale = rbind(locns, locns.test))
  locns.test.orig <- locns.test
  locns.test <- xy2unit(locns.test, locations.scale = rbind(locns, locns.test))
}

# set up mapping of locations to grid
fun <- gp(c(k, k))
map.train <- c(new.mapping(fun, locns))
n <- length(locns[ ,1])
map.test <- c(new.mapping(fun, locns.test))

n.gridded <- rep(0, k * k)
tmp <- table(map.train)
n.gridded[as.numeric(names(tmp))] <- tmp  # same as KtK
n.filled <- n.gridded[n.gridded > 0]
nUnfilled <- sum(n.gridded  == 0)
nFilled <- sum(n.gridded > 0)

# set up mapping from grid to observation locations
indic <- (1:(k * k)) * (n.gridded > 0)
indic <- indic[indic > 0]
invMap.train <- matrix(0, nr = n, nc = nFilled)
for(i in 1:nFilled){
  invMap.train[map.train == indic[i], i] <- 1
}

invMap.train <- t(invMap.train)

filledCells <- (1:(k * k)) * (n.gridded > 0)
filledCells <- filledCells[filledCells > 0]
unfilledCells <- (1:(k * k)) * (n.gridded == 0)
unfilledCells <- unfilledCells[unfilledCells > 0]
if(family == 'poisson'){
  y.filled <- invMap.train %*% y / n.filled
}
if(family == 'binomial'){
  y.filled <- (invMap.train %*% y) / (invMap.train %*% m)
}  
y.gridded <- rep(0, k * k)
y.gridded[filledCells] <- y.filled

# use gam() to get initial fit
dat <- data.frame(y = y, x1 = locns[ ,1], x2 = locns[ ,2])
mod <- gam(y~s(x1, x2, k = min(n - 5, 500)), dat = dat, family = family)  
fullGrid <- expand.gpgrid(fun)

# calculate initial values at grid cells
pred.griddedVec <- predict(mod, newdata = list(x1 = fullGrid[ ,1], x2 = fullGrid[ ,2]),
                           type = 'link') + rnorm(half * half, 0, sqrt(eta2.init))
pred.griddedMat <- matrix(0, nr = k, nc = k)
pred.griddedMat[1:(half), 1:(half)] <- matrix(pred.griddedVec, half, half)
pred.griddedVec <- c(mirror(pred.griddedMat))


#####################################
### set up parameters and processes
#####################################

param <- function(initial = 0, transformFxn = exp, name = NULL, priorParameters = NULL,
                  priorRange = c(-Inf, Inf)){
  # function to initialize parameter objects
  object  <-  list()
  object$name <- name

  object$value <- transformFxn(initial)
  object$baseValue <- initial
  object$p <- length(initial)
  object$transformFxn <- transformFxn

  object$priorParameters <- priorParameters
  object$priorRange <- priorRange

  class(object) <- 'param'
  return(object)
}

identity <- function(x){x}

decide <- function(post.logdiff){
  # Metropolis-Hastings acceptance decision
  n <- length(post.logdiff)
  rdev <- runif(n)
  accepted <- rep(TRUE, n)
  accepted[exp(post.logdiff) <= rdev] <- FALSE
  return(accepted)
}


if(!useJointEta2Sampling){
 # lambda variance component; prior used by Wikle
  eta2 <- param(eta2.init, identity, name = 'eta2', priorParameters = c(.5, .5),
                priorRange = c(0.0001, 100))
} else{
# truncated logNormal is another possibility; use this if jointly sampling eta^2 and
#  augmented lambda values as the code below assumes symmetry on the log scale if doing joint sampling
  eta2 <- param(log(eta2.init), exp, name = 'eta2', priorParameters = c(0, 9),
                priorRange = c(log(.01), 10))
}


# process standard deviation
sigma <- param(log(sd(y)), exp, name = 'sigma', priorParameters = c(0, 9), priorRange = c(-10, 10))
# process range
rho <- param(log(0.05), exp, name = 'rho', priorParameters = c(0, 10),
             priorRange = c(log(.001), log(5)))

# transformed process parameters; these are the parameterization in the MCMC
psi1 <- param(log(sigma$value) + log(rho$value), identity, name = 'psi1',
              priorParameters = c(0, 9), priorRange = c(-15, 15))
psi2 <- param(log(sigma$value) - log(rho$value), identity, name = 'psi2',
              priorParameters = c(0, 9), priorRange = c(-15, 15))

# process mean
mu <- param(linkFxn(mean(y)), identity, name = 'mu', priorParameters = c(0, 10),
            priorRange = c(-Inf, Inf))

# process smoothness
nu <- param(4, identity, name = 'nu', priorParameters = c(0.5, 15), priorRange = c(0.5, 15))

# initialize spatial process
fun <- gp(c(k, k), matern.specdens, c(rho$value, nu$value), sigma$value^2, const.fixed = TRUE)
fun$logdensity <- logdensity(fun)
fun$name <- 'fun'

# 'lambda' process; one lambda value for each grid cell
lambda <- param(pred.griddedVec, identity, name = 'lambda', priorParameters = c(0, 100),
                priorRange = c(-50, 50))
lambda$proposalSD <- rep(0.03, nFilled)
lambda$numProp <- nFilled  # lambda values for grid cells w/o data can be sampled via Gibbs

# initialize spatial process
Gibbs.sample.coeff(fun, lambda$value, eta2$value, mu$value, 1)

########################
### prior and lik fxns
########################

logLik <- function(){
  if(family  == 'poisson'){
    return(sum(dpois(y, invLinkFxn(lambda$value)[map.train], log = TRUE)))
  }
  if(family == 'binomial'){
    return(sum(dbinom(y, m, invLinkFxn(lambda$value)[map.train], log = TRUE)))
  }
}

# logLikelihood summed over observations within each grid cell
logLikFilledVec <- function(){  
  if(family == 'poisson'){
    logLikVec <- dpois(y, invLinkFxn(lambda$value)[map.train], log = TRUE)
  }
  if(family == 'binomial'){
    logLikVec <- dbinom(y, m, invLinkFxn(lambda$value)[map.train], log = TRUE)
  }
  return(invMap.train %*% logLikVec)
}  

lambdaPrior <- function(){
  priorValue <- dnorm(lambda$value, mu$value + fun$process, sqrt(eta2$value), log = TRUE)
  priorValue[lambda$value < lambda$priorRange[1] | lambda$value > lambda$priorRange[2]] <- -Inf
  return(sum(priorValue))
}

# prior for lambda values in cells with data,  summed
lambdaPriorFilled <- function(){
  priorValue <- dnorm(lambda$value[filledCells], mu$value + fun$process[filledCells],
                      sqrt(eta2$value), log = TRUE)
  priorValue[lambda$value[filledCells] < lambda$priorRange[1] | lambda$value[filledCells] >
             lambda$priorRange[2]] <- -Inf
  return(sum(priorValue))
}

# prior for lambda values in cells with data by cell
lambdaPriorFilledVec <- function(){
  priorValue <- dnorm(lambda$value[filledCells], mu$value + fun$process[filledCells],
                      sqrt(eta2$value), log = TRUE)
  priorValue[lambda$value[filledCells] < lambda$priorRange[1] | lambda$value[filledCells] >
             lambda$priorRange[2]] <- -Inf  
  return(priorValue)
}

mu.Gibbs <- function(){
  v <- 1 / ((k * k / eta2$value + 1 / (mu$priorParameters[2]^2)))
  rnorm(1, v * (sum((lambda$value - fun$process)) / eta2$value + mu$priorParameters[1] /
                (mu$priorParameters[2]^2)), sqrt(v))
}

muPrior <- function(){
  if(max(mu$baseValue) > mu$priorRange[2] | min(mu$baseValue) < mu$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(mu$baseValue, mu$priorParameters[1], mu$priorParameters[2], log = TRUE))
  }
}

psi1Prior <- function(){
  if(max(psi1$baseValue) > psi1$priorRange[2] | min(psi1$baseValue) < psi1$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(psi1$baseValue, psi1$priorParameters[1], psi1$priorParameters[2], log = TRUE))
  }
}

psi2Prior <- function(){
  if(max(psi2$baseValue) > psi2$priorRange[2] | min(psi2$baseValue) < psi2$priorRange[1]){
    return(-Inf)
  } else{
    return(dnorm(psi2$baseValue, psi2$priorParameters[1], psi2$priorParameters[2], log = TRUE))
  }
}

eta2.Gibbs <- function(){
  return(1 / rgamma(1, (k * k) / 2 + eta2$priorParameters[1], rate = eta2$priorParameters[2] + 
                  0.5 * sum((lambda$value - mu$value - fun$process)^2)))
}

if(!useJointEta2Sampling){
  eta2Prior <- function(){
    if(max(eta2$baseValue) > eta2$priorRange[2] | min(eta2$baseValue) < eta2$priorRange[1]){
      return(-Inf)
    } else{
      return(-(eta2$priorParameters[1] + 1) * log(eta2$baseValue) - eta2$priorParameters[2] /
             eta2$baseValue)
    }
  }
} else{
# use with logNormal prior on eta^2
  eta2Prior <- function(){
    if(max(eta2$baseValue) > eta2$priorRange[2] | min(eta2$baseValue) < eta2$priorRange[1]){
      return(-Inf)
    } else{
      return(dnorm(eta2$baseValue, eta2$priorParameters[1], eta2$priorParameters[2], log = TRUE))
    }
  }
}


funPrior <- function(){
  logdensity(fun)
}

psi1$proposalSD <- 0.03
psi2$proposalSD <- 0.03
# use with joint sampling of eta^2 and augmented lambda
if(useJointEta2Sampling){
  eta2$proposalSD <- 0.03
}

totaln <- nBurnin + nIts
cat('Starting iterations: ',date(),'\n')

for(it in 1:totaln){

  # sample coefficients
  Gibbs.sample.coeff(fun, lambda$value, eta2$value, mu$value, 1)

  # lambda sample

  # filled locations - Metropolis sample individually for each location
  savedValue <- lambda$value[filledCells]
  lpValue <- logLikFilledVec() + lambdaPriorFilledVec()
  accepted <- rep(TRUE, nFilled)
  proposed <- rnorm(nFilled, savedValue, lambda$proposalSD)
  lambda$value[filledCells] <- proposed
  lpProp <- logLikFilledVec() + lambdaPriorFilledVec()
  accepted <- decide(lpProp - lpValue)
  accepted[proposed < lambda$priorRange[1] | proposed > lambda$priorRange[2]] <- FALSE
  lambda$value[filledCells][accepted == FALSE] <- savedValue[accepted == FALSE]
    
  # unfilled locations - Gibbs sample
  lambda$value[unfilledCells] <- rnorm(nUnfilled, (mu$value + fun$process)[unfilledCells],
                                       sqrt(eta2$value))


  # sample lambda variance component either using Gibbs sample or with joint sample with
  #    augmented lambda process
  if(!useJointEta2Sampling){
    eta2$value <- eta2.Gibbs()
  } else{
   # if doing joint eta^2, augmented lambda sampling:
    eta2Save <- eta2
    lpValue <- eta2Prior() + lambdaPriorFilled()
    eta2$baseValue <- rnorm(1, eta2$baseValue, eta2$proposalSD) 
    eta2$value <- eta2$transformFxn(eta2$baseValue)
    hastingsVal <- 0
    lpProp <- eta2Prior() + lambdaPriorFilled()
    accepted <- decide(lpProp - lpValue + hastingsVal)
    if(accepted){
      lambda$value[unfilledCells] <- rnorm(nUnfilled, (mu$value + fun$process)[unfilledCells],
          sqrt(eta2$value))  # sample lambda values at unfilled cells if joint proposal accepted
    } else{
      eta2 <- eta2Save
    }
  }

  
  # sample first process parameter jointly with coefficients using deterministic move of coefficients
  psi1Save <- psi1
  lpValue <- psi1Prior() + logLik()
  psi1$baseValue <- rnorm(1, psi1$baseValue, psi1$proposalSD)
  psi1$value <- psi1$transformFxn(psi1$baseValue)
  funSave <- copy(fun)
  rhoSave <- rho
  sigmaSave <- sigma
  sigma$value <- exp((psi1$value + psi2$value) / 2)   # express in terms of standard parameters
  rho$value <- exp((psi1$value - psi2$value) / 2)
  change.param(fun, c(rho$value, nu$value), sigma$value^2)
  fun$coeff <- fun$coeff * sqrt(fun$variances / funSave$variances)  # deterministic shift of coefficients
  updateprocess(fun)
  lambdaSave <- lambda
  lambda$value <- lambda$value - funSave$process + fun$process
      # deterministic shift of lambda process values
  lpProp <- psi1Prior() + logLik()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    lambda <- lambdaSave
    rho <- rhoSave
    sigma <- sigmaSave
    psi1 <- psi1Save
    copy(funSave, fun)
  }

  # sample second process parameter jointly with coefficients using deterministic move of coefficients
  psi2Save <- psi2
  lpValue <- psi2Prior() + logLik()
  psi2$baseValue <- rnorm(1, psi2$baseValue, psi2$proposalSD)
  psi2$value <- psi2$transformFxn(psi2$baseValue)
  funSave <- copy(fun)
  rhoSave <- rho
  sigmaSave <- sigma
  sigma$value <- exp((psi1$value + psi2$value) / 2)  # express in terms of standard parameters
  rho$value <- exp((psi1$value - psi2$value) / 2)
  change.param(fun, c(rho$value, nu$value), sigma$value^2)
  fun$coeff <- fun$coeff * sqrt(fun$variances / funSave$variances)  # deterministic shift of coefficients
  updateprocess(fun)
  lambdaSave <- lambda
  lambda$value <- lambda$value - funSave$process + fun$process
                                        # deterministic shift of lambda process values 
  lpProp <- psi2Prior() + logLik()
  accepted <- decide(lpProp - lpValue)
  if(!accepted){
    lambda <- lambdaSave
    rho <- rhoSave
    sigma <- sigmaSave
    psi2 <- psi2Save
    copy(funSave, fun)
  }


  # sample process mean
  mu$value <- mu.Gibbs()
  
  # calculate process values at training, test, and all grid locations
  # note that this uses smoothest version of process based on GP
  f.train <- mu$value + predict(fun, mapping = map.train)
         # note that we use error-free version of lambda
  f.test <- mu$value + predict(fun, mapping = map.test)
  f.grid <- mu$value + predict(fun)
}
